from os.path import join
from typing import Optional, Tuple, List

class XfoilPolar():
    name: 'str' = None
    numpnl: 'int' = None
    re: Optional['float'] = None
    mach: Optional['float'] = None
    ncrit: 'float' = None
    xtrftop: 'float' = None
    xtrfbot: 'float' = None
    alpha: List['float'] = None
    cl: List['float'] = None
    cd: List['float'] = None
    cdp: List['float'] = None
    cm: List['float'] = None
    trtop: List['float'] = None
    trbot: List['float'] = None
    def __init__(self, name: 'str', numpnl: 'int') -> None:
        self.name = name
        self.numpnl = numpnl
    def read_polar(self, filepath: str) -> None:
        self.alpha = []
        self.cl = []
        self.cd = []
        self.cdp = []
        self.cm = []
        self.trtop = []
        self.trbot = []
        read = False
        with open(filepath, 'rt') as file:
            for line in file:
                line = line.rstrip('\n')
                if line.strip() == '':
                    continue
                if line[0:23] == ' Calculated polar for: ':
                    self.name = line[23:].strip()
                if line[0:8] == ' xtrf = ':
                    self.xtrftop = float(line[8:15])
                    self.xtrfbot = float(line[27:34])
                if line[0:8] == ' Mach = ':
                    self.mach = float(line[8:15])
                    self.re = float(line[25:38].replace(' ', ''))
                    self.ncrit = float(line[51:58])
                if line[0:8] == '   alpha':
                    read = True
                    continue
                if read:
                    if line[0:8] != '  ------':
                        self.alpha.append(float(line[0:8]))
                        self.cl.append(float(line[8:17]))
                        self.cd.append(float(line[17:27]))
                        self.cdp.append(float(line[27:37]))
                        self.cm.append(float(line[37:46]))
                        self.trtop.append(float(line[46:55]))
                        self.trbot.append(float(line[55:64]))
    @property
    def clocd(self) -> Optional[List['float']]:
        if self.cl is not None and self.cd is not None:
            clocdval = [cli/cdi for cli, cdi in zip(self.cl, self.cd)]
            return clocdval
    def plot_polar(self, xaxis='cd', yaxis='cl', ax=None, *args, **kwargs):
        figsize = kwargs.get('figsize', (12, 8))
        if ax is None:
            from matplotlib.pyplot import figure
            fig = figure(figsize=figsize)
            ax = fig.gca()
            title = r'Polar plot for $M = {:g}$ and $Re = {:.12g}$'
            ax.set_title(title.format(self.mach, self.re))
            grid = kwargs.get('grid', True)
            ax.grid(grid)
            xlabel = self.get_label(xaxis)
            ylabel = self.get_label(yaxis)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
        label = r''
        if self.re is not None:
            label += r'$Re = {:.12g}$'.format(self.re)
        if self.re is not None and self.mach is not None:
            label += r'; '
        if self.mach is not None:
            label += r'$M = {:g}$'.format(self.mach)
        xvalue = self.get_value(xaxis)
        yvalue = self.get_value(yaxis)
        ax.plot(xvalue, yvalue, *args, label=label)
        return ax
    def get_label(self, var: 'str') -> 'str':
        if var == 'alpha':
            label = r'$\alpha$'
        elif var == 'cl':
            label = '$C_l$'
        elif var == 'cd':
            label = '$C_d$'
        elif var == 'cdp':
            label = '$C_{dp}$'
        elif var == 'cm':
            label = '$C_m$'
        elif var == 'toptr':
            label = r'$Top X Transition$'
        elif var == 'th':
            label = r'$Bottom X Transition$'
        elif var == 'clocd':
            label = r'$\frac{c_l}{c_d}$'
        else:
            raise ValueError(f'{var:s} does not exist in XfoilPolar.')
        return label
    def get_value(self, var: 'str') -> List['float']:
        if var == 'alpha':
            value = self.alpha
        elif var == 'cl':
            value = self.cl
        elif var == 'cd':
            value = self.cd
        elif var == 'cdp':
            value = self.cdp
        elif var == 'cm':
            value = self.cm
        elif var == 'toptr':
            value = self.trtop
        elif var == 'th':
            value = self.trbot
        elif var == 'clocd':
            value = self.clocd
        else:
            raise ValueError(f'{var:s} does not exist in XfoilPolar.')
        return value
    def __repr__(self) -> str:
        return f'<pyxfoil.XfoilPolar {self.name:s}>'

def write_polar_session(name: 'str', datfilepath: 'str', numpnl: 'int',
                        almin: 'float', almax: 'float', alint: 'float',
                        mach: Optional['float']=None,
                        re: Optional['float']=None,
                        ppar: Optional['int']=None) -> Tuple['str', 'str']:
    from pyxfoil import workdir
    polname = name.replace(' ', '_') + f'_{numpnl:d}'
    if mach is not None:
        polname += f'_{mach:g}'
    if re is not None:
        polname += f'_{re:.12g}'
    filepath = join(workdir, polname)
    sesfilepath = f'{filepath:s}.ses'
    polfilepath = f'{filepath:s}.pol'
    with open(sesfilepath, 'wt') as file:
        file.write('load {:s}\n'.format(datfilepath))
        if ppar is not None:
            file.write('ppar\n')
            file.write('n {:d}\n'.format(ppar))
            file.write('\n')
            file.write('\n')
        file.write('oper\n')
        if mach is not None:
            file.write('mach {:g}\n'.format(mach))
        if re is not None:
            file.write('visc {:.12g}\n'.format(re))
        file.write('pacc\n')
        file.write('{:s}\n'.format(polfilepath))
        file.write('\n')
        file.write('aseq\n')
        file.write('{:g}\n'.format(almin))
        file.write('{:g}\n'.format(almax))
        file.write('{:g}\n'.format(alint))
        file.write('pacc\n')
        if mach is not None:
            file.write('mach 0.0\n')
        if re is not None:
            file.write('visc\n')
        file.write('\n')
        file.write('quit\n')
    return sesfilepath, polfilepath



from os.path import join
from typing import Tuple, Optional, List

class XfoilResult():
    name: 'str' = None
    numpnl: 'int' = None
    alpha: 'float' = None
    re: Optional['float'] = None
    mach: Optional['float'] = None
    s: List['float'] = None
    x: List['float'] = None
    y: List['float'] = None
    ue: List['float'] = None
    ds: List['float'] = None
    th: List['float'] = None
    cf: List['float'] = None
    h: List['float'] = None
    resfile: 'str' = None
    _cp: List['float'] = None
    def __init__(self, name: 'str', numpnl: 'int') -> None:
        self.name = name
        self.numpnl = numpnl
    def set_param(self, alpha: 'float', mach: 'float', re: 'float') -> None:
        self.alpha = alpha
        self.mach = mach
        self.re = re
    def read_result(self, resfile: 'str') -> None:
        with open(resfile, 'rt') as f:
            self.s = []
            self.x = []
            self.y = []
            self.ue = []
            self.ds = []
            self.th = []
            self.cf = []
            self.h = []
            for line in f:
                line = line.rstrip('\n')
                if line != '':
                    if line[0] != '#':
                        self.s.append(float(line[1:11]))
                        self.x.append(float(line[11:20]))
                        self.y.append(float(line[20:29]))
                        self.ue.append(float(line[29:38]))
                        self.ds.append(float(line[38:48]))
                        self.th.append(float(line[48:58]))
                        self.cf.append(float(line[58:68]))
                        self.h.append(float(line[68:78]))
        self._cp = None
    @property
    def cp(self) -> List['float']:
        if self._cp is None:
            self._cp = [1-uei**2 for uei in self.ue]
        return self._cp
    def plot_result(self, xaxis='x', yaxis='ue', ax=None, *args, **kwargs):
        figsize = kwargs.get('figsize', (12, 8))
        if ax is None:
            from matplotlib.pyplot import figure
            fig = figure(figsize=figsize)
            ax = fig.gca()
            grid = kwargs.get('grid', True)
            ax.grid(grid)
            xlabel = self.get_label(xaxis)
            ylabel = self.get_label(yaxis)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            title = r'Result plot for $\alpha = {:g}$'.format(self.alpha)
            if self.re is not None:
                title += r' and $Re = {:.12g}$'.format(self.re)
            if self.mach is not None:
                title += r' and $M = {:g}$'.format(self.mach)
            ax.set_title(title)
            if yaxis == 'cp':
                ax.invert_yaxis()
        label = r'$\alpha = {:g}$'.format(self.alpha)
        if self.re is not None:
            label += r'; $Re = {:.12g}$'.format(self.re)
        if self.mach is not None:
            label += r'; $M = {:g}$'.format(self.mach)
        xvalue = self.get_value(xaxis)
        yvalue = self.get_value(yaxis)
        ax.plot(xvalue, yvalue, *args, label=label)
        return ax
    def result(self, var: 'str', correct: 'bool'=False) -> List['float']:
        res = self.get_value(var)
        if correct:
            if var == 's':
                offset = max(res[:self.numpnl+1])
                val = [offset-resi for resi in res[:self.numpnl+1]]
            else:
                val = [resi for resi in res[:self.numpnl+1]]
            val.reverse()
            val = val + res[self.numpnl+1:]
        else:
            val = res.copy()
        return val
    def get_label(self, var: 'str') -> 'str':
        if var == 'x':
            label = '$x$'
        elif var == 'y':
            label = '$y$'
        elif var == 's':
            label = '$s$'
        elif var == 'ue':
            label = '$u_e$'
        elif var == 'cp':
            label = '$c_p$'
        elif var == 'ds':
            label = r'$\delta^*$'
        elif var == 'th':
            label = r'$\theta$'
        elif var == 'h':
            label = '$h$'
        elif var == 'cf':
            label = '$c_f$'
        else:
            raise ValueError(f'{var:s} does not exist in XfoilResult.')
        return label
    def get_value(self, var: 'str') -> List['float']:
        if var == 'x':
            value = self.x
        elif var == 'y':
            value = self.y
        elif var == 's':
            value = self.s
        elif var == 'ue':
            value = self.ue
        elif var == 'cp':
            value = self.cp
        elif var == 'ds':
            value = self.ds
        elif var == 'th':
            value = self.th
        elif var == 'h':
            value = self.h
        elif var == 'cf':
            value = self.cf
        else:
            raise ValueError(f'{var:s} does not exist in XfoilResult.')
        return value
    def __repr__(self) -> str:
        return f'<pyxfoil.XfoilResult {self.name:s}>'

def write_result_session(name: 'str', datfilepath: 'str', numpnl: 'int',
                         alpha: 'float', mach: Optional['float']=None,
                         re: Optional['float']=None,
                         ppar: Optional['int']=None) -> Tuple['str', 'str']:
    from pyxfoil import workdir
    resname = name.replace(' ', '_')
    resname += f'_{numpnl:d}_{alpha:g}'
    if mach is not None:
        resname += f'_{mach:g}'
    if re is not None:
        resname += f'_{re:.12g}'
    filepath = join(workdir, resname)
    sesfilepath = f'{filepath:s}.ses'
    resfilepath = f'{filepath:s}.res'
    with open(sesfilepath, 'wt') as file:
        file.write('load {:s}\n'.format(datfilepath))
        if ppar is not None:
            file.write('ppar\n')
            file.write('n {:d}\n'.format(ppar))
            file.write('\n')
            file.write('\n')
        file.write('oper\n')
        if mach is not None:
            file.write('mach {:g}\n'.format(mach))
        if re is not None:
            file.write('visc {:.12g}\n'.format(re))
        file.write('alfa {:g}\n'.format(alpha))
        file.write('dump {:s}\n'.format(resfilepath))
        if mach is not None:
            file.write('mach 0.0\n')
        if re is not None:
            file.write('visc\n')
        file.write('\n')
        file.write('quit\n')
    return sesfilepath, resfilepath



from os import system, remove
from os.path import isfile, join, split
from typing import List, Dict
from .xfoilresult import XfoilResult, write_result_session
from .xfoilpolar import XfoilPolar, write_polar_session
from matplotlib.pyplot import figure

class Xfoil():
    name: 'str' = None
    ppar: 'int' = None
    x: List['float'] = None
    y: List['float'] = None
    results: Dict['str', 'XfoilResult'] = None
    polars: Dict['str', 'XfoilPolar'] = None
    _area: 'float' = None
    def __init__(self, name: 'str') -> None:
        self.name = name
    def points_from_dat(self, datfile: 'str') -> None:
        self.x = []
        self.y = []
        with open(datfile, 'rt') as file:
            for i, line in enumerate(file):
                line = line.rstrip('\n')
                if i == 0:
                    self.name = line.strip()
                else:
                    split = line.split()
                    if len(split) == 2:
                        x = float(split[0])
                        y = float(split[1])
                        self.x.append(x)
                        self.y.append(y)
        self._area = None
    def set_points(self, x: List['float'], y: List['float']) -> None:
        self.x = x
        self.y = y
        self._area = None
    def set_ppar(self, ppar: 'int') -> None:
        self.ppar = ppar
    @property
    def area(self) -> 'float':
        if self._area is None:
            self._area = 0.0
            for i in range(1, len(self.x)):
                self._area += self.x[i]*self.y[i-1]-self.y[i]*self.x[i-1]
            if self.x[0] != self.x[-1] or self.y[0] != self.y[-1]:
                self._area += self.x[0]*self.y[-1]-self.y[0]*self.x[-1]
        return self._area
    def write_dat(self) -> 'str':
        from pyxfoil import workdir
        datname = self.name.replace(' ', '_')
        filepath = join(workdir, datname)
        datfilepath = f'{filepath:s}.dat'
        num = len(self.x)
        if self.area >= 0:
            order = range(num-1, -1, -1)
        else:
            order = range(num)
        frmstr = '  {:11.6f} {:11.6f}\n'
        with open(datfilepath, 'wt') as file:
            file.write(self.name+'\n')
            for i in order:
                file.write(frmstr.format(self.x[i], self.y[i]))
        return datfilepath
    def run_result(self, alfa: 'float', re: 'float'=None,
                   mach: 'float'=None) -> 'XfoilResult':
        from pyxfoil import xfoilexe
        datfilepath = self.write_dat()
        numpnl = len(self.x) - 1
        sesfilepath, resfilepath = write_result_session(self.name, datfilepath, numpnl,
                                                        alfa, mach=mach, re=re,
                                                        ppar=self.ppar)
        if isfile(resfilepath):
            remove(resfilepath)
        system('{:s} < {:s}'.format(xfoilexe, sesfilepath))
        res = split(resfilepath)[1]
        res = res.replace('.res', '')
        result = XfoilResult(res, numpnl)
        result.set_param(alfa, mach, re)
        result.read_result(resfilepath)
        if self.results is None:
            self.results = {}
        self.results[res] = result
        return result
    def run_polar(self, almin: 'float', almax: 'float', alint: 'float',
                  re: 'float'=None, mach: 'float'=None) -> 'XfoilPolar':
        from pyxfoil import xfoilexe
        datfilepath = self.write_dat()
        numpnl = len(self.x) - 1
        sesfilepath, polfilepath = write_polar_session(self.name, datfilepath,
                                                       numpnl, almin, almax, alint,
                                                       mach=mach, re=re,
                                                       ppar=self.ppar)
        if isfile(polfilepath):
            remove(polfilepath)
        system('{:s} < {:s}'.format(xfoilexe, sesfilepath))
        pol: 'str' = split(polfilepath)[1]
        pol = pol.replace('.pol', '')
        polar = XfoilPolar(pol, numpnl)
        polar.read_polar(polfilepath)
        if self.polars is None:
            self.polars = {}
        self.polars[pol] = polar
        return polar
    def plot_profile(self, *args, **kwargs):
        grid = kwargs.get('grid', True)
        aspect = kwargs.get('aspect', 'equal')
        figsize = kwargs.get('figsize', (12, 8))
        fig = figure(figsize=figsize)
        ax = fig.gca()
        ax.plot(self.x, self.y, *args, **kwargs)
        ax.grid(grid)
        ax.set_aspect(aspect)
        ax.set_title('Airfoil profile for {:}.'.format(self.name))
        return ax



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = '''
    Run XFoil from within Python and work with output results.
    '''

setup(
    name="pyxfoil",
    version="0.0.2",
    author="Xero64",
    author_email="xero64@gmail.com",
    description="Run XFoil from within Python and work with output results.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Xero64/pyxfoil",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)
